/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * parallel version of matrix multiply
 *
 * 
 *  +-   -+ +-       -+     +-       -+
 *  |     | |         |     |         |
 *  |  A  | |    B    |  =  |    C    | 
 *  | MxP | |   PxN   |     |   MxN   |
 *  |     | +-       -+     |         |
 *  |     |                 |         |
 *  +-   -+                 +-       -+
 *
 *  C[i][j] =   SUM     A[i][k] * B[k][j]
 *            k=1 to P
 *
 *  
 * 31 Jul 92 (was) initial version
 * 21 Aug 92 (was) changed to run under PC, RC or EC
 *  5 Oct 92 (was) changed to run stand-alone
 *  6 Oct 92 (was) converted to new midway naming
 * 21 Oct 92 (was) maximize performance
 * xx Oct 92 (mjz): various fiddling
 *  3 nov 92 (mjz): do 't' for touch; are verifying... sigh.
 * 14 Oct 93 (mjz): update for new ATM + header files
 */
#define DEBUG       0     /* define > 0 for debugging output */
/*#define STAND_ALONE 0*/    /* =1 to compile out multiprocessor code */
#define ZERO_RESULTS 0
#define VERIFY       1
#define WAIT		0
#define MM		4 /* MM Version number */

#ifdef MAPPED
#define GETTIMEOFDAY my_gettimeofday
#else
#define GETTIMEOFDAY gettimeofday
#endif

#include <stdio.h>

#if STAND_ALONE == 0
#include "midway.h"
#endif

#include <math.h>

#define KERNEL 1		/* HACK to resolve type conflict. WAS 5Oct92 */
#include <sys/time.h>
#undef KERNEL



#if DEBUG > 0
#define PRINTF printf
#else
#define PRINTF (void)
#endif

#if STAND_ALONE > 0
#define midway_lock_alloc()  0
#define midway_lock_acquire (void)
#define midway_lock_release (void)
#define midway_cross_barrier (void)
#define midway_flush_barrier (void)
#define midway_barrier_alloc() 0
#define midway_create_thread (void)
#define midway_bind_synch_object (void)
#define mid_comm_print_stats()
#define mid_lock_print_stats()
#define mid_xpager_print_stats()
#define mid_paged_print_stats()
#define midway_uxpass_remote()
#define midway_exit(X)
#define StatusMsg(X)
#define botch(X) printf(X); printf("\n");
#define timer_pause_ms (void)
#define shmalloc(size,cache) malloc(size)
#define midway_lock_t int
#define midway_barrier_t int
#define LOCK_RO 0
#define LOCK_RW 0
#define shared 
#define private
#endif

/*
 * Define program limits 
 */
#define DEFAULT_MATRIX_SIZE      100
#define DEFAULT_CACHE_LINE_SIZE 1024
#define MAX_PROCS                 32

/*
 * useful defines
 */
#define NULL 0
#define USECS_PER_SEC  1000000

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

/*
 * If you change these, check out BlockBind below!
 */
#define INDEX(row,col) (((row)*N)+(col))
#define A_INDEX(row,col) (((row)*N)+(col))
#define B_INDEX(row,col) (((row)*N)+(col))
#define C_INDEX(row,col) (((row)*N)+(col))
#define a_index(row,col) (((row)*n)+(col)) /* use register vars for globals */
#define b_index(row,col) (((row)*n)+(col))
#define c_index(row,col) (((row)*n)+(col))

#define MODEL_EC    0		/* Entry consistency */
#define MODEL_RC    1		/* Release consistency */
#define MODEL_PC    2		/* Processor consistency */
#define MODEL_UNI   3           /* Stand alone consistency */


/*
 * Define matrix data type
 */
typedef float data_t;
#define print_data(x) printf("%f", x);
#define PRINT_DATA(x) PRINTF("%f", x);

/*
 * Define per-thread data
 */
typedef struct {
    int			startingRow;
    int			endingRow;
    int			startingCol;
    int			endingCol;
    midway_lock_t 	InputLock;  /* A matrix values */
    midway_lock_t	OutputLock; /* C matrix values */
} per_thread_t;

/*
 * Declare shared variables
 */

shared static data_t *A;            /* ptr to input MxP matrix */
shared static data_t *B;            /* ptr to input PxN matrix */
shared static data_t *C;            /* ptr to output MxN matrix */
shared static int N;                /* array size */
shared static int NProcs;           /* number of processes */

shared static per_thread_t PerThreadData[ MAX_PROCS ];
#if 0
shared static int Start[MAX_PROCS+1]; /* starting point for each processor */
#endif

shared static midway_lock_t  MasterLock;   /* guards readonly data */
#if 0
shared static midway_lock_t *InputLock;    /* guards rows in array A */
shared static midway_lock_t *OutputLock;   /* mutex array guarding rows in array C */
#endif
shared static midway_barrier_t WaitPoint;  /* thread barrier */
shared static midway_lock_t stat_lock;	   /* for stats at end */

shared static int Model;            /* consistency model */
shared static int touch;	    /* touch pages before acquiring stuff */
shared static int useIdentity;	/* set B = I, so can easily verify C. */
shared static int useBarrier;


/*
 * Declare external routines 
 */
#if STAND_ALONE == 0
extern void StatusMsg (char *);   /* send a msg to the status server for debugging */
#endif
extern char *optarg;


#define MASTER_THREAD_ID 0  /* Id used by master thread */



/*
**
** NOTE: Locking is handled as follows:
**       EC - locks all shared data R/O, locks each output row in R/W
**            uses a barrier to synchronize task termination
**       RC - locks each output row (in R/W)
**            uses a barrier to synchronize task termination
**       PC - doesn't use locks,
**            uses a barrier to synchronize task termination
**
** This application does not require the use of locks on the output matrix C
** since it is partitioned by thread id. The locks are used solely for testing
** and debugging of midway code and may be removed for actual performance
** measurements.  WAS 21Aug92
**
*/

/*
 * Routine to perform actual multiplication
 */
void MMultiply (int myId)
{
    per_thread_t *threadP;
    int i,j;
    int n;
    int model;
    data_t result;
    shared data_t *a, *b, *c;
    shared data_t *aValuesP, *bValuesP, *cValuesP;
    shared data_t *aEndP;
    int sRow, eRow, sCol, eCol;
    struct timeval startlock, endlock, startcompute, endcompute;
    struct timeval locktime, lockbartime, computetime;
    struct timeval startTouch, endTouch, touchTime;
    struct timeval startunlk, endunlk, unlkTime;
    void SubtractTime(struct timeval s, struct timeval e, struct timeval *r);
    volatile register data_t zorch;
    int b_is_identity;

#ifdef MAPPED
    if (myId)
#if 0
	my_gettimeofday(&startlock, 0); /* gotta prime pump on slaves */
#else
        /* With PagedBased, the timer base & init vars are overwritten,
	   so we must do it by hand */
        setup_mapped_timer();
#endif
#endif MAPPED

    /* cache starting & ending row information */
    threadP = &PerThreadData[ myId ];
    sRow = threadP->startingRow;
    eRow = threadP->endingRow;
    sCol = threadP->startingCol;
    eCol = threadP->endingCol;
    
    PRINTF("%d: Thread %d waiting to start %d to %d C %d to %d, Bar=0x%x NProcs= %d \n", 
	   myId, myId, sRow, eRow, sCol, eCol, WaitPoint, NProcs);

    /* cache all static vars */
    n = N;
    a = A;
    b = B;
    c = C;
    model = Model;
    b_is_identity = useIdentity;

    if ((touch != 0) & (myId != 0)) {
	int step = 4096 /* page size */ / sizeof(data_t);

	/* need to touch only rows used for non-ec */

	my_gettimeofday(&startTouch, 0);

	/* A ARRAY */
	aValuesP = &A[ A_INDEX(sRow,0) ];
	aEndP = &A[ A_INDEX(eRow,N-1) ];

	for (; aValuesP <= aEndP; aValuesP+=step)
	        zorch = *aValuesP;


	/* niether B nor C arrays guaranteed to use whole rows */
	j = step / n; /* == pgsz DIV szof(row) == row inc to hit pgs */
	j = j * n;		/* now back to addresses */

	/* B ARRAY */
	aValuesP = &B[ B_INDEX(0,sCol) ];
	aEndP = &B[ B_INDEX(N-1,eCol) ];

	for (; aValuesP <= aEndP; aValuesP+=j)
	        zorch = *aValuesP;

	/* touch last in case not aligned */
	zorch = *aEndP;


	/* C ARRAY */
	aValuesP = &C[ C_INDEX(sRow,sCol) ];
	aEndP = &C[ C_INDEX(eRow,eCol) ];

	for (; aValuesP <= aEndP; aValuesP+=j)
	        zorch = *aValuesP;

	/* touch last in case not aligned */
	zorch = *aEndP;

	my_gettimeofday(&endTouch, 0);
    }

    /* wait for everyone else before starting */
    /* not proc 0 so it can be counted on main. */
    if (myId) {
	midway_cross_barrier (WaitPoint, NProcs);
    }

    
    GETTIMEOFDAY(&startlock, 0);
    /* Acquire lock on ReadOnly data */
    if (model == MODEL_EC) {
	PRINTF("%d: Thread about to acquire master lock = %d (0x%x)\n", 
	       myId, MasterLock, &MasterLock);
	midway_lock_acquire (MasterLock, LOCK_RO);
	
	/* Now acquire lock guarding rows in matrix A */
	midway_lock_acquire (threadP->InputLock, LOCK_RO);
    }

    /* acquire exclusive lock on output rows */
#ifdef oldWay
    if ((model == MODEL_EC) || (model == MODEL_RC))
#else
    if (!useBarrier && (model == MODEL_EC))
#endif
	midway_lock_acquire (threadP->OutputLock, LOCK_RW);

    GETTIMEOFDAY(&endlock, 0);

    /* wait for all to grab thier data */
    midway_cross_barrier(WaitPoint, NProcs);

    GETTIMEOFDAY(&startcompute, 0);

    /* compute row in result based on processor id */
    for (i = sRow; i <= eRow; i++) {
	cValuesP = &c[ c_index(i,sCol) ]; /* not guaranteed to be consec. */
    
	PRINTF("%d: Thread computing row %d [c %d-%d] at 0x%x (0x%x)\n", 
	       myId, i, sCol, eCol, &c[ c_index(i,0) ], cValuesP);
    
	/* compute each element in row */
	for (j = sCol; j <= eCol; j++) {

	    aValuesP = &a[ a_index(i,0) ];
	    aEndP = aValuesP + n;   /* A's stride is N */
	    bValuesP = &b[ b_index(0,j) ];
	    result = 0;
	    do {
		result += *aValuesP * *bValuesP;
		aValuesP++;
		bValuesP += n;   /* B's stride is N */
	    } while (aValuesP < aEndP);

	    *cValuesP = result;
	    cValuesP++;

	    PRINTF("%d:                  row %d col %d = ", 
		   myId, i, j);
	    PRINT_DATA (result);
	    PRINTF("\n");

#if VERIFY > 0
	    /* VERIFY DATA */
	    if (b_is_identity && (a[ A_INDEX(i,j) ] != c[ C_INDEX(i,j) ])) {
		printf("Thread Computed Invalid result at 0x%x,  C[%d][%d] = ",
		       &C[ C_INDEX(i,j) ], i, j);
		print_data( C[ C_INDEX(i,j) ]);
		printf(" should be ");
		print_data( A[ A_INDEX(i,j) ]);
		printf("\n");
		printf("A=0x%x B=0x%x C=0x%x, N=%d myId=%d, SR=%d ER=%d SC=%d EC=%d, result=%f\n",
		       A, B, C, N, myId, sRow, eRow, sCol, eCol, result);
		printf("a=0x%x b=0x%x c=0x%x, n=%d aValP=0x%x cValP=0x%x\n",
		       a, b, c, n, aValuesP, cValuesP);
		botch("application");
	    }
#endif
	}
	
    }

    GETTIMEOFDAY(&endcompute, 0);
    
    PRINTF("%d: Thread done computing\n", myId);

#ifdef oldWay
    GETTIMEOFDAY(&startunlk, 0);
    /* release lock on output rows */
    if ((model == MODEL_EC) || (model == MODEL_RC))
	midway_lock_release (threadP->OutputLock);
    GETTIMEOFDAY(&endunlk, 0);

    /* Release master lock before exiting */
    if (model == MODEL_EC) {
	midway_lock_release (threadP->InputLock);
	midway_lock_release (MasterLock);
    }
    
    PRINTF("%d: Thread waiting at termination barrier\n", myId);

    /* Wait for everyone else to finish */
    midway_cross_barrier (WaitPoint, NProcs);
#else
    /* release lock on output rows */
    if (!useBarrier && ((model == MODEL_EC)))
	midway_lock_release (threadP->OutputLock);

    PRINTF("%d: Thread waiting at termination barrier\n", myId);

    /* Wait for everyone else to finish */
    GETTIMEOFDAY(&startunlk, 0);
    if (useBarrier) {
	midway_flush_barrier (WaitPoint, NProcs);
    } else {
	midway_cross_barrier (WaitPoint, NProcs);
    }
    GETTIMEOFDAY(&endunlk, 0);

    /* Release master lock before exiting */
    if (model == MODEL_EC) {
	midway_lock_release (threadP->InputLock);
	midway_lock_release (MasterLock);
    }
    
#endif

    PRINTF("%d: Thread done\n", myId);

    /* Slaves wait again to print usage stats */
    if (myId != 0) {
	midway_cross_barrier (WaitPoint, NProcs);
	midway_lock_acquire (stat_lock, LOCK_RW);
	/* Print midway usage stats */
	mid_comm_print_stats();
	mid_lock_print_stats();
	mid_xpager_print_stats();
	mid_paged_print_stats();
    }
    SubtractTime(startlock, endlock, &locktime);
    SubtractTime(startcompute, endcompute, &computetime);
    SubtractTime(endlock, startcompute, &lockbartime);
    SubtractTime(startunlk, endunlk, &unlkTime);
	/* for some reason, uxpass is local now??? */
	midway_uxpass_remote();
    printf("Proc %d: locktime %ld.%06ld sec, lbartime %ld.%06ld s, computetime %ld.%06ld sec unlk %ld.%06lds\n",
	   myId,
	   locktime.tv_sec, locktime.tv_usec,
	   lockbartime.tv_sec, lockbartime.tv_usec,
	   computetime.tv_sec, computetime.tv_usec,
	   unlkTime.tv_sec, unlkTime.tv_usec);
    if ((touch != 0) && (myId !=0)) {
	SubtractTime(startTouch, endTouch, &touchTime);
	printf("Proc %d: touchtime %ld.%06ld sec\n",
	       myId, touchTime.tv_sec, touchTime.tv_usec);
    }
    fflush(stdout);
    StatusMsg("at end of compute");
    if (myId != 0) {
	midway_lock_release (stat_lock);
	midway_cross_barrier (WaitPoint, NProcs);
    }
}


void SubtractTime (struct timeval start, 
		   struct timeval end, 
		   struct timeval *diff)
{
    if (end.tv_usec < start.tv_usec) {
	end.tv_usec += USECS_PER_SEC;
	end.tv_sec--;
    }
    
    if (end.tv_sec < start.tv_sec)
	printf("Invalid time!\n");
    
    diff->tv_sec  = end.tv_sec - start.tv_sec;
    diff->tv_usec = end.tv_usec - start.tv_usec;
}

void ProgramDone ()
{
    printf("\nmm: terminating\n\n");
    exit(0);
}

/*
 * Following binds a block of a matrix delineated by the starting and
 * ending rows and columns to a lock or barrier (don't let midway_lock_t
 * fool you (although I wish I could specify this better...).
 */
static void BlockBind (midway_lock_t X, shared data_t *M,
		       int sRow, int eRow, int sCol, int eCol)
{
    int r, l;

    if (Model != MODEL_EC)
	return;

    l = (eCol - sCol) + 1;
    for (r = sRow; r <= eRow; r++) {
	PRINTF("binding SV=0x%x row=%d sCol=%d eCol=%d addr=0x%x bytes=%d\n",
	       X, r, sCol, eCol, 
	       &M[INDEX(r, sCol)], l*sizeof(data_t));
	midway_bind_synch_object (X,
				  &M[INDEX(r, sCol)], l * sizeof(data_t));
    } /* for rows */
} /* BlockBind */


/*
 * main
 */
main (argc, argv)
     int argc;
     char **argv;
{
    char c;
    int i, j;
    int len;
    int aSize, aCache;
    int bSize, bCache;
    int cSize, cCache;
    int lockSize, lockCache;
    struct timeval startTime, stopTime, computeTime, initTime;
    struct timeval diffTime1, diffTime2, diffTime3, diffTime4, diffTime5;
    struct timezone timeZone;
    int work;
    int error;
    int p, split, part, diff, mid;
    int sRow, eRow;
    per_thread_t *threadP;

#ifdef MAPPED
    my_gettimeofday(&startTime, 0); /* prime pump */
#endif

#if WAIT
    printf("mm: matrix multiplication\n");
    printf("sleeping for %d seconds\n",WAIT);
    timer_pause_ms (WAIT * 1000);    /* time in millieseconds */
    printf("resuming mm\n");
#endif

    GETTIMEOFDAY (&initTime, &timeZone);

    printf("mm: argc=%d argv=", argc);
    for (i = 0; i < argc; i++) printf(" \"%s\"", argv[i]);
    printf("\n");


    /* Initialize command line variables */
    N = DEFAULT_MATRIX_SIZE;
    NProcs = 1;
    aCache = DEFAULT_CACHE_LINE_SIZE;
    bCache = DEFAULT_CACHE_LINE_SIZE;
    cCache = DEFAULT_CACHE_LINE_SIZE;
    Model = MODEL_EC;
    work = 100;
    touch = 0;
    useBarrier = 0;
    
    /* set vars from command line */
    while ((c = getopt(argc, argv, "Bitp:n:c:m:w:")) != EOF)
	switch (c)
	    {
	    case 'p':
		NProcs = atoi(optarg);
		break;
	    case 'n':
		N = atoi(optarg);
		break;
	    case 'c':
		aCache = bCache = cCache = atoi(optarg);
		break;
	    case 'm':
		if (!strcmp(optarg,"EC")) {
		    Model = MODEL_EC;
		} else if (!strcmp(optarg,"RC")) {
		    Model = MODEL_RC;
		} else if (!strcmp(optarg,"PC")) {
		    Model = MODEL_PC;
		} else 
		    printf("Invalid consistency model: \"%s\"\n", optarg);
		
		printf("Consistency model set to %s\n", 
		       (Model == MODEL_EC) ? "EC" : 
		       (Model == MODEL_RC) ? "RC" : 
		       (Model == MODEL_PC) ? "PC" : "UNKNOWN");
		break;
	    case 'w':
		work = atoi(optarg);
		break;
	    case 't':
		touch++;	/* set touch the data flag */
		break;
	    case 'i':
		useIdentity = 1; /* set B == I. */
		break;
	    case 'B':
		useBarrier = 1;	/* for returning C matrix data */
		break;
	    default:
		printf("%s: Illegal flag %c, usage: -t -p # -n # -w# -c # -m(EC,RC,PC)\n", 
		       argv[0], c);
		ProgramDone();
		break;
	    }
    
    if (N <= 0) {
	printf("Done: matrix size of 0\n");
	ProgramDone();
    }
    if (NProcs <= 0) {
	printf("Done: no threads assigned to perform work\n");
	ProgramDone();
    }
    if (NProcs > MAX_PROCS) {
	printf("Done: Number of processes (%d) greater than max (%d)\n",
	       NProcs, MAX_PROCS);
	ProgramDone();
    }
    if (NProcs == 1)
	Model = MODEL_UNI;


    printf("Multiplying (%d,%d) * (%d,%d) = (%d,%d) using %d processes\n",
	   N, N, N, N, N, N, NProcs);
    if (useIdentity) {
	printf("A * B = C, B will be the identity matrix.\n");
    }
    
    /*
     * Allocate work to processors first
     */
    /* do we split into left and right halves? */
    if ((NProcs >= 4) && ((NProcs % 2) == 0))
	split = TRUE;
    else
	split = FALSE;

    if (split) {
	part = N / (NProcs / 2); 	/* rows/processor */
	diff = N - (part * (NProcs/2));	/* left over, if not even split */
	/* -1, for index origin of 0. */
	mid = (N / 2)-1;		/* split row where? */
    } else {
	part = N / NProcs;		/* rows/processor */
	diff = N - (part * NProcs); 	/* left over, if not even split */
    }

    sRow = 0;
    for (p = 0; p < NProcs; p++) {
	
	threadP = &PerThreadData[p];
	threadP->startingRow = sRow;
	/* allocate extra rows left over one each among the first processors */
	if (split)
	    eRow = sRow + (((p/2) < diff) ? part : part - 1);
	else
	    eRow = sRow + ((p < diff) ? part : part - 1);
	threadP->endingRow = eRow;
	if (eRow > N) {
	    printf("Too many processes %d for %d rows of work\n", NProcs, N);
	    exit(1);
	}
	if (split) {
	    if ((p % 2) == 0) {
		threadP->startingCol = 0;
		threadP->endingCol = mid;
	    } else {
		threadP->startingCol = mid + 1;
		threadP->endingCol = N-1;
		sRow = eRow + 1;
	    }
	} else {
	    threadP->startingCol = 0;
	    threadP->endingCol = N-1;
	    sRow = eRow + 1;
	}
	PRINTF("Proc %d: rows %d to %d, columns %d to %d (inclusive)\n",
	       p, threadP->startingRow, threadP->endingRow,
	       threadP->startingCol, threadP->endingCol);
    } /* allocating work to processors */

    
    /* Allocate matrices from shared storage */
    aSize = sizeof(data_t) * N * N;
    if ((aCache < sizeof(data_t)) || (aCache > aSize))
	aCache = aSize;
    A = (shared data_t *) shmalloc (aSize, aCache);
    
    bSize = sizeof(data_t) * N * N;
    if ((bCache < sizeof(data_t)) || (bCache > bSize))
	bCache = bSize;
    B = (shared data_t *) shmalloc (bSize, bCache);
    
    cSize = sizeof(data_t) * N * N;
    { int maxCache = sizeof(data_t) * N;
      if (split) {
	  if (maxCache % 2)
	      maxCache = sizeof(data_t); /* can't split in two. sigh. */
	  else
	      maxCache = maxCache / 2;
      }
      if ((cCache > maxCache) || (cCache > cSize) || (cCache < sizeof(data_t)))
	  cCache = maxCache;

      /* Check that the cache line size evenly divides the rows */
      i = sizeof(data_t) * N;
      if (split)
	  i = i/2;
      if ((i % cCache) != 0) {
	  printf("Cache line size in output matrix set to %d bytes\n", i);
	  cCache = i;
      }
    }
    C = (shared data_t *) shmalloc (cSize, cCache);
    
    if (A == NULL || B == NULL || C == NULL) {
	printf("Done: Insufficient space for matrices (%d bytes)\n",
	       aSize + bSize + cSize);
	ProgramDone();
    }
    
    printf("matrix A at 0x%x, size=%d, element=%d, cacheline=%d\n", 
	   A, aSize, sizeof(data_t), aCache);
    printf("matrix B at 0x%x, size=%d, element=%d, cacheline=%d\n", 
	   B, bSize, sizeof(data_t), bCache);
    printf("matrix C at 0x%x, size=%d, element=%d, cacheline=%d\n", 
	   C, cSize, sizeof(data_t), cCache);

    
    /* Initialize input matrices */
    for (i = 0; i < N; i++)
	for (j = 0; j < N; j++)
	    A[ A_INDEX(i,j) ] = (i*N) + j;
    if (useIdentity) {
	for (i = 0; i < N; i++)
	    for (j = 0; j < N; j++)
		B[ B_INDEX(i,j) ] = (i == j) ? 1 : 0;
    } else {
	for (i = 0; i < N; i++)
	    for (j = 0; j < N; j++)
		B[ B_INDEX(i,j) ] = 1;
    }

#if ZERO_RESULTS > 0
    for (i = 0; i < N; i++)
	for (j = 0; j < N; j++)
	    C[ C_INDEX(i,j) ] = 0;
#endif


#if 0
    /* Partition work among the slaves by rows in output matrix
     * Since master doesn't wait for data to arrive
     * let master do twice as much work.
     */
    Start[0] = 0;
    Start[NProcs] = N;
    if (NProcs > 1) {
	if (work < 10) work = 10;
	if (work > 1000) work = 1000;
	j = N / ((NProcs * work) / 100);
	if (j < 1) {
	    printf("Insufficient work for %d processes\n", NProcs);
	    ProgramDone();
	}
	for (i = NProcs-1; i > 0; i--)
	    Start[i] = Start[i+1] - j;
	
	if (work != 100) {
	    printf("Work %d/100: Output rows partition: Master %d, Slave%s %d",
		   work, Start[0], (NProcs > 2) ? "s" : "", Start[1]);
	    for (i = 2; i < NProcs; i++)
		printf(" %d", Start[i]);
	    printf("\n");
	}
    }
#endif

    if (touch) {
	volatile register data_t zorch;

	if (Model == MODEL_RC) {
	    /* 9nov92 mjz: for RC, do not touch, as the slaves will touch indirectly */
	    /* touch only our rows */
	    for (i = PerThreadData[0].startingRow; 
		 i <= PerThreadData[0].endingRow; i++) 
		for (j = PerThreadData[0].startingCol; 
		     j <= PerThreadData[0].endingCol; j++)
		    zorch = C[ C_INDEX(i,j) ];
	} else {		/* model not rc  */
	    for (i = 0; i < N; i++)
		for (j = 0; j < N; j++)
	            zorch = C[ C_INDEX(i,j) ];

	} /* model not rc */
    } /* touch */
	

    /* Allocate locks for readonly input matrices */
    if (Model == MODEL_EC) {

	MasterLock = midway_lock_alloc();
#if 0
	midway_bind_synch_object (MasterLock, B, bSize);
#endif

	/* bind per-processor array to the Master Lock */
	midway_bind_synch_object (MasterLock, PerThreadData, 
				  sizeof(PerThreadData));
    
	for (i = 0; i < NProcs; i++) {
	    /*
	     * allocate input locks
	     * bind rows of A, and columns of B to each lock.
	     * Double binding is dangerous, but we will allow it
	     * for the input matricies... since only one site writes it.
	     */
	    threadP = &PerThreadData[i];
	    threadP->InputLock = midway_lock_alloc();
	    BlockBind(threadP->InputLock, A,
		      threadP->startingRow, threadP->endingRow,
		      0, N-1);
	    BlockBind(threadP->InputLock, B,
		      0, N-1,
		      threadP->startingCol, threadP->endingCol);
	}
    }

    if (!useBarrier) {
	if ((Model == MODEL_EC) || (Model == MODEL_RC)) {
	    for (i = 0; i < NProcs; i++) {
		threadP = &PerThreadData[i];
		threadP->OutputLock = midway_lock_alloc();
		if (Model == MODEL_EC) {
		    BlockBind(threadP->OutputLock, C,
			      threadP->startingRow, threadP->endingRow,
			      threadP->startingCol, threadP->endingCol);
		}
	    }
	}
    }

    /* Allocate a barrier to synchronize threads */
    WaitPoint = midway_barrier_alloc ();

    /* Bind output array to barrier */
    if (useBarrier && (Model == MODEL_EC)) {
	midway_bind_synch_object (WaitPoint, C, cSize);
    }

    /* Allocate a lock to synchronize stats so we don't blow out
       netmsgserver (hack) */
    stat_lock = midway_lock_alloc ();

    /*
     * HACK: 8nov92mjz; updateq stuff does not purge if already acquired
     * via a page fault.  start with null update q.
     */
    if (Model == MODEL_RC) {
	midway_lock_acquire(PerThreadData[0].OutputLock, LOCK_RW);
	midway_lock_release(PerThreadData[0].OutputLock);
    }
  
    /* Spawn slave threads to perform work for us */
    for (i = 1; i < NProcs; i++) {
	PRINTF("%d: Master starting slave %d\n",
	       MASTER_THREAD_ID, i);
	midway_create_thread (MMultiply, i);
    }

    if (NProcs > 1)
	printf("Master will compute %d rows, each slave %d\n", 
	       PerThreadData[0].endingRow - PerThreadData[0].startingRow+1,
	       PerThreadData[1].endingRow - PerThreadData[0].startingRow+1);

    printf("Ready to execute, using %s consistency model\n",   
	   (Model == MODEL_EC) ? "ENTRY" : 
	   (Model == MODEL_RC) ? "RELEASE" : 
	   (Model == MODEL_PC) ? "PROCESSOR" :
	   (Model == MODEL_UNI) ? "UNIPROCESSOR" : "INVALID");


    /* wait for all slaves here.  This will include 'touch' time, if any */
    midway_cross_barrier (WaitPoint, NProcs);


    /* start timer */
    GETTIMEOFDAY (&startTime, &timeZone);
    
    /* Master needs to perform work as well */
    MMultiply ( MASTER_THREAD_ID );

    /* stop timer */
    GETTIMEOFDAY (&computeTime, &timeZone); /* compute time here is bogus */
  
    /* Master must obtain data back from slaves before terminating clock */
    if (!useBarrier && (Model == MODEL_EC)) {
	midway_lock_acquire (MasterLock, LOCK_RO);
	for (i = 0; i < NProcs; i++)
	    midway_lock_acquire (PerThreadData[i].OutputLock, LOCK_RO);
    }
    
    /* stop timer */
    GETTIMEOFDAY (&stopTime, &timeZone);
  
    SubtractTime (initTime, startTime, &diffTime1);
    SubtractTime (startTime, computeTime, &diffTime2);
    SubtractTime (computeTime, stopTime, &diffTime3);
    SubtractTime (startTime, stopTime, &diffTime4);
    SubtractTime (initTime, stopTime, &diffTime5);
    printf ("\nDone.\n");
    printf("(Init=%ld.%06ld), Compute=%ld.%06ld, xfer=%ld.%06ld, total time = %ld.%06ld (%ld.%06ld) secs.\n\n",
	    diffTime1.tv_sec, diffTime1.tv_usec,
	    diffTime2.tv_sec, diffTime2.tv_usec,
	    diffTime3.tv_sec, diffTime3.tv_usec,
	    diffTime4.tv_sec, diffTime4.tv_usec,
	    diffTime5.tv_sec, diffTime5.tv_usec);
    
    /* lock input array as well */
    if (Model == MODEL_EC) {
	for (i = 0; i < NProcs; i++)
	    midway_lock_acquire (PerThreadData[i].InputLock, LOCK_RO);
    }
    
    /* verify results for square matrices only */
    error = 0;
    if (useIdentity) {
	for (i = 0; i < N; i++) {
	    for (j = 0; j < N; j++) {
		if (A[ A_INDEX(i,j) ] != C[ C_INDEX(i,j) ]) {
		    printf("Invalid result at 0x%x,  C[%d][%d] = ",
			   &C[ C_INDEX(i,j) ], i, j);
		    print_data( C[ C_INDEX(i,j) ]);
		    printf(" should be ");
		    print_data( A[ A_INDEX(i,j) ]);
		    printf("\n");
		    
		    error++;
		}
	    }
	}
    } /* useIdentity*/

    if (Model == MODEL_EC) {
	for (i = 0; i < NProcs; i++) {
	    if (!useBarrier)
		midway_lock_release (PerThreadData[i].OutputLock);
	    midway_lock_release (PerThreadData[i].InputLock);
	}
	midway_lock_release (MasterLock);
    }

    if (error) {
	printf("\nFinal Result Matrix:\n");
	for (i = 0; i < N; i++) {
	    printf("%5d:", i);
	    for (j = 0; j < N; j++) {
		printf(" ");
		print_data( C[ C_INDEX(i,j) ]);
	    }
	    printf("\n");
	}
    }

    /* Wait again to print usage stats */
    midway_lock_acquire (stat_lock, LOCK_RW);
    midway_cross_barrier (WaitPoint, NProcs);

    mid_comm_print_stats();
    mid_lock_print_stats();
    mid_xpager_print_stats();
    mid_paged_print_stats();

    midway_lock_release (stat_lock);
    midway_cross_barrier(WaitPoint, NProcs);

    midway_exit (error);
}



/*
 * Set stuff up so that Brian's indentation is preserved when editing with
 * Gnu-emacs:
 *
 * Local Variables:
 * c-indent-level: 4
 * c-continued-statement-offset: 4
 * c-label-offset: -4
 * End:
 *
 * c-brace-offset
 */
