/* 
 * Midway Distributed Shared Memory System
 * Copyright (c) 1994 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 */
/*
 * timing.c: a file of utilities for gathering timing statistics
 *
 * 27 Jul 92 (mjz): from cs412 ta stuff last year
 * 26 Oct 92 (mjz): use mapped timer [pmax, need grp=kmem or uid=0]
 */
#include "timing.h"
#ifdef MAPPED			/* use mapped timer facility */
#include <stdio.h>
#include <mach.h>
#include <mach/message.h>
#include <mach/error.h>
#include <device/device_types.h>
#include <mach/time_value.h>

static mapped_init = 0;
volatile struct mapped_time_value *mtime;

setup_mapped_timer()
{
    long secs, usecs;
    kern_return_t rc;
    mach_port_t device_server_port, device_port, pager = MACH_PORT_NULL;

    /* gak backdoor.  need kmem group/root to doo this. */
    device_server_port = task_by_pid(-2);
    rc = device_open(device_server_port,D_READ,"time",&device_port);
    if (rc != KERN_SUCCESS) {
/*      printf("line:%d\n", __LINE__);*/
      printf("device_open error: %s\n", mach_error_string(rc));
      exit(1);
    }

    rc = device_map(device_port, VM_PROT_READ,
		    0, vm_page_size, &pager, 0);
    if (rc != D_SUCCESS) {
/*      printf("line:%d\n", __LINE__);*/
      printf("device_map error: %s\n", mach_error_string(rc));
      exit(1);
    }
    if (pager == MACH_PORT_NULL) {
/*      printf("line:%d\n", __LINE__); */
      printf("device_map (no pager) error: %s\n", mach_error_string(rc));
      exit(1);
    }


    rc = vm_map(mach_task_self(),
		(vm_offset_t *) &mtime, sizeof *mtime, 0, TRUE,
		pager, 0, FALSE,
		VM_PROT_READ, VM_PROT_READ, VM_INHERIT_NONE);
    if (rc != D_SUCCESS) {
/*      printf("line:%d\n", __LINE__);*/
      printf("vm_map error: %s\n", mach_error_string(rc));
      exit(1);
    }

    rc = mach_port_deallocate(mach_task_self(), pager);
    if (rc != KERN_SUCCESS) {
/*      printf("line:%d\n", __LINE__);*/
      printf("mach_port_deallocate error: %s\n", mach_error_string(rc));
      exit(1);
    }

    mapped_init = 1;

} /* setup_mapped_timer */

my_gettimeofday(tv_t *tp, struct timezone* tzp)
{
  if (!mapped_init) {
    setup_mapped_timer();
  }

  do {				/* see mach/time_value.h */
    tp->tv_sec = mtime->seconds;
    tp->tv_usec = mtime->microseconds;
  } while (tp->tv_sec != mtime->check_seconds);
  /* we don't do timezones. */
}/* my gettimeofday */ 
#else /* not mapped */
my_gettimeofday(tv_t *tp, struct timezone* tzp)
{
  return(gettimeofday(tp, tzp));
}
#endif /* MAPPED */

/*
 * compute the difference between two timevals
 */
void
time_diff(start, fin, result)
     tv_t *start, *fin, *result;
     /* subtract start from fin and place in result */
{
  result->tv_sec = fin->tv_sec - start->tv_sec;
  result->tv_usec = fin->tv_usec - start->tv_usec;
  /* account for underflow in usecs */
  if (result->tv_usec < 0) {
    result->tv_usec += 1000000;	/* add in a second */
    result->tv_sec -= 1;	/* and subtract from here */
  }

  return;
} /* time_diff */


#define MIL	1000000
#define FMIL    1000000.0

void
time_init( ts )
     timestat_t *ts;
{
  ts->sum = 0.0;
  ts->sumsq = 0.0;
  ts->mean = 0.0;
  ts->var = 0.0;
  ts->sdev = 0.0;
  ts->count = 0;
} /* time_init */

/*
 * time_add: add the difference of two timevals to a timestat
 *
 * returns
 *	difference between the two timevals
 */
double
time_add( start, end, ts, num )
     tv_t *start;
     tv_t *end;
     timestat_t *ts;
     int num;			/* num of iterations btwn start & end */
{
  tv_t diff;			/* the difference */
  double d;			/* difference as a double */

  time_diff(start, end, &diff);	/* compute difference as a timeval */
  d = TV_TO_DOUBLE(diff);
  /* ? do we want to do this here? */
  d *= FMIL / (double) num;	/* convert to microsec */
  ts->sum += d;			/* keep running sum */
  ts->sumsq += (d * d);		/* and sum of squares */
  ts->count++;			/* one more value here */

  return(d);
} /* time_add */

/*
 * time_stat: compute statistics on a timestat struct
 */
void
time_stat( ts )
     timestat_t *ts;
{
  double sqrt();

  ts->mean = ts->sum / ((double)(ts->count));
  if (ts->count == 1)
    return;			/* rest make no sense */
  ts->var = (ts->sumsq - ((ts->sum * ts->sum) / ((double)(ts->count)))) /
    (double)(ts->count - 1);
  ts->sdev = sqrt(ts->var);
} /* time_stat */

#ifdef EXAMPLE_USE
/* to see example, do:
 * cc -o timing timing.c -lm -DEXAMPLE_USE
 */

#include <stdio.h>

main()
{
  tv_t start, fin, result, loop_only, final_result;
  double dresult, dloop, dfinal;
  int i,j;
  int pid;
  int onum;			/* overall number of iterations */
  int num;			/* number of iterations to get timing */
  timestat_t pcall;
  timestat_t loop;

  /* for testing, make these numbers small (so I don't fall asleep using
   * Mach 3.0... */
  num = /*MIL*/ 100000;		/* do one million times */
  onum =/*50*/10;		/* fifty good for statistical significance */
  /* ?? maybe not */


  time_init( &loop );

  for(j=0; j<onum; j++) {
    /* measure just loop */
    gettimeofday(&start, NOZONE);
    for (i=0; i<num; i++) {
      pid = i;
    }
    gettimeofday(&fin, NOZONE);
    dloop = time_add(&start, &fin, &loop, num);
    time_diff(&start, &fin, &loop_only);
    printf("Loop %d times in %d.%06.6ld sec [%f usec for overhead]\n",
	   num, loop_only.tv_sec, loop_only.tv_usec, dloop);
  }

  /* now compute mean and std dev */
  time_stat(&loop, onum);
  printf("--> mean %lf var %lf stdev %lf sum %lf sumsq %lf [microseconds]\n",
	 loop.mean, loop.var, loop.sdev, loop.sum, loop.sumsq);


  putc('\n', stdout);

  time_init( &pcall );
  for (j=0; j<onum; j++) {	/* do ten times to gather stats */
    gettimeofday(&start, NOZONE);
    for (i=0; i<num; i++) {
      pid = getpid();
    }
    gettimeofday(&fin, NOZONE);
    dresult = time_add(&start, &fin, &pcall, num);
    time_diff(&start, &fin, &result); /* for printf */
    printf("getpid %d times in %d.%06.6ld sec [%f usec/op]\n",
	   num, result.tv_sec, result.tv_usec, dresult);
  }
  
  time_stat( &pcall, onum ); /* compute pcall stats */
  printf("--> mean %lf var %lf stdev %lf sum %lf sumsq %lf [microseconds]\n",
	 pcall.mean, pcall.var, pcall.sdev, pcall.sum, pcall.sumsq);

  putc('\n',stdout);

  dfinal = pcall.mean - loop.mean;
  printf("Average system call time without loop overhead: %lf microseconds.\n",
	 dfinal);

} /* main */

#endif EXAMPLE_USE
